import logging
import copy
import pymacnet.messages
from datetime import datetime
from .cycler_interface import CyclerInterface

logger = logging.getLogger(__name__)


class ChannelInterface(CyclerInterface):
    """
    Class for controlling Maccor Cycler using MacNet.
    """

    def __init__(self, config: dict):
        """
        Creates a ChannelInterface class instance.

        Parameters
        ----------
        config : dict
            A configuration dictionary. Must contain the following keys:
            - `server_ip` - The IP address of the Maccor server. Use 127.0.0.1 if running on the same machine as the server.
            - `json_msg_port` - The port to communicate through with JSON messages. Default set to 57570.
            - `bin_msg_port` - The port to communicate through with binary messages. Default set to 57560.
            - `msg_buffer_size_bytes` - How big of a message buffer to use for sending/receiving messages. A minimum of 1024 bytes is recommended. 
            - `channel` - The channel to be targeted for all operations.
            - `test_name` - The test name to be used for any tests started. If left blank, Maccor will generate a unique random name for any started tests. Note that Maccor requires unique test names for each test.
            - `test_procedure` - The test procedure to be used, if starting a test with a procedure. Not needed with direct control.
            - `c_rate_ah` - The capacity value to be referenced when setting "C" values within the Maccor schedule. Units of amp-hours. Ignored if not used anywhere in the test.
            - `v_max_safety_limit_v` - Upper voltage safety limit for the channel. Units of volts.
            - `v_min_safety_limit_v` - Lower voltage safety limit for the channel. Units of volts.
            - `i_max_safety_limit_a` - Upper current safety limit for the channel. Units of amps.
            - `i_min_safety_limit_a` - Lower current safety limit for the channel. Units of amps.
            - `v_max_v` - Upper voltage limit used for charge/CV limits. Units of volts. Only used with direct control.
            - `v_min_v` - Lower voltage limit used for discharge limit. Units of volts. Only used with direct control.
            - `data_record_time_s` - How often data points are taken during direct control tests. Zero turns off. Used only for direct control.
            - `data_record_voltage_delta_vbys` - The dV/dt at which data points are taken during direct control tests. Zero disables. Used only for direct control.
            - `data_record_current_delta_abys` - The dI/dt at which data points are taken during direct control tests. Zero disables. Used only for direct control.
        """
        self.__channel = config['channel']
        self.__config = config

        assert (self.__verify_config())
        super().__init__(config=config)

    def get_channel_number(self):
        '''
        Returns the channel number for the class instance.
        '''
        return self.__channel

    def read_channel_status(self) -> dict:
        """
        Method to read the status of the channel defined in the config.

        Returns
        -------
        status : dict
            A dictionary detailing the status of the channel. Returns None if there is an issue.
        """
        return super().read_channel_status(channel=(self.__channel))

    def read_aux(self) -> list:
        """
        Reads the auxiliary readings for the channel specified in the config.

        Returns
        -------
        aux_readings : list
            A list of the auxiliary readings.
        """

        msg_outgoing_dict = copy.deepcopy(pymacnet.messages.tx_read_aux_msg)
        msg_outgoing_dict['params']['Chan'] = self.__channel

        aux_readings = self._send_receive_json_msg(msg_outgoing_dict)
        if aux_readings:
            return aux_readings['result']['AuxValues']
        else:
            logger.error("Failed to read channel aux values!")
            return None

    def reset_channel(self) -> bool:
        """
        Resets the channel. Note this will stop any actively running test on the target channel.

        Returns
        -------
        success : bool
            True of False based on whether the test was started or not.
        """
        msg_outgoing_dict = copy.deepcopy(
            pymacnet.messages.tx_reset_channel_msg
        )
        msg_outgoing_dict['params']['Chan'] = self.__channel

        reply = self._send_receive_json_msg(msg_outgoing_dict)
        if reply:
            if reply['result']['Result'] == 'OK':
                success = True
            else:
                logger.error("Failed reset channel! Channel did not reset!")
                success = False
        else:
            logger.error("Failed reset channel! Did not receive reply!")
            success = False

        return success

    def set_channel_variable(self, var_num=1, var_value=0) -> bool:
        """
        Sets test variables on the target channel. See the "Variables" section in the 
        Maccor manual for more details about how to use these in tests.

        Parameters
        ----------
        var_num : int
            Value between 1 and 16, depending on which variable to set.
        var_value : float
            Value to set the variable to.

        Returns
        -------
        success : bool
            True of False based on whether or not the variable value was set.
        """
        msg_outgoing_dict = copy.deepcopy(
            pymacnet.messages.tx_set_variable_msg)
        msg_outgoing_dict['params']['Chan'] = self.__channel
        msg_outgoing_dict['params']['VarNum'] = var_num
        msg_outgoing_dict['params']['Value'] = var_value

        # Check to make variable was set
        reply = self._send_receive_json_msg(msg_outgoing_dict)
        if reply:
            try:
                assert ((reply['result']['Chan']) == self.__config['channel'])
                assert (reply['result']['Result'] == 'OK')
            except:
                logger.error("Variable not set!")
                return False
        else:
            logger.error(
                "Failed to receive reply message when trying to set variable!")
            return False

        return True

    def start_test_with_procedure(self) -> bool:
        """
        Starts the test on the channel and with the procedure specified in the passed config. 
        Note that it will not start a test if the channel is current running a test.

        Returns
        -------
        success : bool
            True of False based on whether the test was started or not
        """
        msg_outgoing_dict = copy.deepcopy(
            pymacnet.messages.tx_start_test_with_procedure_msg)
        msg_outgoing_dict['params']['Chan'] = self.__channel
        msg_outgoing_dict['params']['ProcName'] = self.__config['test_procedure']
        msg_outgoing_dict['params']['Crate'] = self.__config['c_rate_ah']
        msg_outgoing_dict['params']['Comment'] = "Started with pymacnet at " + \
            str(datetime.timestamp(datetime.now()))

        # If test name is not specified then start test with a random test
        if not self.__config['test_name']:
            msg_outgoing_dict['params']['TestName'] = 'Random'
        else:
            msg_outgoing_dict['params']['TestName'] = self.__config['test_name']

        # Check the status of the channel before we try to start the test.]
        status = self.read_channel_status()
        if status:
            if pymacnet.messages.status_dictionary[status['Stat']] == 'Completed':
                self.reset_channel()
        else:
            logger.error("Cannot read channel status!")
            return False

        # Set the safety limits.
        if not self.__set_channel_safety_limits():
            logger.error("Failed to set channel safety limits!")
            return False

        # Start the test.
        response = self._send_receive_json_msg(msg_outgoing_dict)
        if response:
            if response['result']['Result'] != 'OK':
                logger.error("Error starting test! Comment from Maccor: " +
                             response['result']['Result'])
                return False
            else:
                return True
        else:
            logger.error(
                "Failed to get message response when trying to start test!")
            return False

    def start_test_with_direct_control(self) -> bool:
        """
        Starts a test to be manually controlled with direct output on the channel specified in the config.

        Returns
        -------
        success : bool
            True of False based on whether the test was started or not.
        """

        msg_outgoing_dict = copy.deepcopy(
            pymacnet.messages.tx_start_test_with_direct_control_msg)
        msg_outgoing_dict['params']['Chan'] = self.__channel
        msg_outgoing_dict['params']['DataTime'] = self.__config['data_record_time_s']
        msg_outgoing_dict['params']['DataV'] = self.__config['data_record_voltage_delta_vbys']
        msg_outgoing_dict['params']['DataI'] = self.__config['data_record_current_delta_abys']
        # Make sure the start current is always zero.
        msg_outgoing_dict['params']['Current'] = 0
        # Set to something within range so it's not disabled.
        msg_outgoing_dict['params']['Voltage'] = self.__config['v_max_v']
        # TODO: FIX having a weird issue where I can't set the mode to rest.
        msg_outgoing_dict['params']['ChMode'] = "C"

        # If test name is not specified then start test with a random test
        if not self.__config['test_name']:
            msg_outgoing_dict['params']['TestName'] = "Random"
        else:
            msg_outgoing_dict['params']['TestName'] = self.__config['test_name']

        # Check the status of the channel before we try to start the test.
        status = self.read_channel_status()
        if status:
            if pymacnet.messages.status_dictionary[status['Stat']] == 'Completed':
                self.reset_channel()
        else:
            logger.error("Cannot read channel status!")
            return False

        # Set the safety limits.
        if not self.__set_channel_safety_limits():
            logger.error("Failed to set channel safety limits!")
            return False

        # Start the test
        response = self._send_receive_json_msg(msg_outgoing_dict)
        if response:
            if response['result']['Result'] != 'OK':
                logger.error("Error starting test! Comment from Maccor: " +
                             response['result']['Result'])
                return False
            else:
                return True
        else:
            logger.error(
                "Failed to get message response when trying to start test!")
            return False

    def set_direct_mode_output(self, current_a, voltage_v=4900) -> bool:
        """
        Sets the current/voltage output on the channel specified on in the config. Note that the
        test must have been started with the start_test_direct_control method for this to work.

        For discharging: Indicated with a negative sign on current. A lower voltage limit is set based on the
            `v_min_v` value set in the config.

        For charging: If only `current_a` is passed, the cycler will charge at the this current until commanded 
            otherwise or until the upper voltage safety limit is hit. If a `voltage_v` argument is passed in addition to `current_a`, then the cycler will do a CCCV 
            charge with the requested voltage.


        Returns
        -------
        success : bool
            True of False based on whether the values were set or not.
        """

        # Determine mode and voltage limits based on sign of passed current.
        if current_a == 0:
            set_current_a = 0
            set_voltage_v = 0
            mode = "R"
        elif current_a > 0:
            set_current_a = current_a
            set_voltage_v = voltage_v
            mode = "C"
        elif current_a < 0:
            set_current_a = abs(current_a)
            set_voltage_v = self.__config['v_min_v']
            mode = "D"
        else:
            logger.error("Undefined state is set direct control output!")
            return False

        # Determine range based on the magnitude of the current
        if set_current_a == 0:
            # Don't bother changing current range if we're just setting to zero.
            current_range = 4
        elif set_current_a < 0.000150:
            current_range = 1
        elif set_current_a < 0.005:
            current_range = 2
        elif set_current_a < 0.150:
            current_range = 3
        else:
            current_range = 4

        msg_outgoing_dict = copy.deepcopy(
            pymacnet.messages.tx_set_direct_output_msg)
        msg_outgoing_dict['params']['Chan'] = self.__channel
        msg_outgoing_dict['params']['Current'] = set_current_a
        msg_outgoing_dict['params']['Voltage'] = set_voltage_v
        msg_outgoing_dict['params']['ChMode'] = mode
        msg_outgoing_dict['params']['CurrentRange'] = current_range

        # Send the rest command using the caveman method as JSON cannot set rest currently (10/4/2022)
        if mode == "R":
            return self._send_direct_output_rest_bin_msg(msg_outgoing_dict)
        else:
            # Send message and make sure response indicates values were accepted.
            response = self._send_receive_json_msg(msg_outgoing_dict)
            if response:
                if response['result']['Result'] == "OK":
                    return True
                else:
                    logger.error(
                        "Error Setting output! Message : " + str(response))
                    return False
            else:
                logger.error(
                    "Failed to get message response when trying to set output!")
                return False

    def __set_channel_safety_limits(self) -> bool:
        """
        Sets channel safety limits on the channel specified in the config. 

        Returns
        -------
        success : bool
            Returns True/False based on whether the safety limits were set correctly.
        """
        msg_outgoing_dict = copy.deepcopy(
            pymacnet.messages.tx_set_safety_limits_msg
        )
        msg_outgoing_dict['params']['Chan'] = self.__channel
        msg_outgoing_dict['params']['VSafeMax'] = self.__config['v_max_safety_limit_v']
        msg_outgoing_dict['params']['VSafeMin'] = self.__config['v_min_safety_limit_v']
        msg_outgoing_dict['params']['ISafeChg'] = self.__config['i_max_safety_limit_a']
        msg_outgoing_dict['params']['ISafeDis'] = self.__config['i_min_safety_limit_a']
        msg_outgoing_dict['params']['PBatSafeChg'] = self.__config['power_safety_limit_chg_w']
        msg_outgoing_dict['params']['PBatSafeDis'] = self.__config['power_safety_limit_dsg_w']

        # Check to make sure all safety limits were set correctly
        reply = self._send_receive_json_msg(msg_outgoing_dict)
        if reply:
            try:
                assert ((reply['result']['Chan']) == self.__config['channel'])
                assert (abs(reply['result']['VSafeMax'] -
                        self.__config['v_max_safety_limit_v']) < 0.001)
                assert (abs(reply['result']['VSafeMin'] -
                        self.__config['v_min_safety_limit_v']) < 0.001)
                assert (abs(reply['result']['ISafeChg'] -
                        self.__config['i_max_safety_limit_a']) < 0.001)
                assert (abs(reply['result']['ISafeDis'] -
                        self.__config['i_min_safety_limit_a']) < 0.001)
                assert (abs(reply['result']['PBatSafeChg'] -
                        self.__config['power_safety_limit_chg_w']) < 0.001)
                assert (abs(reply['result']['PBatSafeDis'] -
                        self.__config['power_safety_limit_dsg_w']) < 0.001)
            except:
                logger.error(
                    "Set safety limits do not match sent safety limits! Message response: " + str(reply))
                return False
        else:
            logger.error(
                "Failed to receive reply message when trying to set safety limits!")
            return False

        return True

    def __verify_config(self) -> bool:
        """
        Verifies that the config passed on construction is valid.

        Returns
        --------------------------
        success : bool
            True or False based on whether the config passed at construction is valid.
        """
        success = False

        required_config_keys = ['channel',
                                'test_name',
                                'test_procedure',
                                'v_max_safety_limit_v',
                                'v_min_safety_limit_v',
                                'i_max_safety_limit_a',
                                'i_min_safety_limit_a',
                                'power_safety_limit_chg_w',
                                'power_safety_limit_dsg_w',
                                'v_max_v',
                                'v_min_v',
                                'c_rate_ah',
                                'data_record_time_s',
                                'data_record_voltage_delta_vbys',
                                'data_record_current_delta_abys',
                                'server_ip',
                                'json_msg_port',
                                'bin_msg_port']

        missing_keyes = False
        for key in required_config_keys:
            if key not in self.__config:
                logger.error("Missing key from config! Missing : " + key)
                missing_keyes = True

        if not missing_keyes:
            success = True

        return success



import logging
import copy
import socket
import logging
import json
import struct
import pymacnet.messages

logger = logging.getLogger(__name__)


class CyclerInterface():
    """
    Class for interfacing with Maccor Cycler using MacNet.
    """

    def __init__(self, config: dict):
        """
        Creates a CyclerInterface class instance. 

        Parameters
        ----------
        config : dict
            A configuration dictionary. Must contain the following keys:
            - `server_ip` - The IP address of the Maccor server. Use 127.0.0.1 if running on the same machine as the server.
            - `json_msg_port` - The port to communicate through with JSON messages. Default set to 57570.
            - `bin_msg_port` - The port to communicate through with binary messages. Default set to 57560.
            - `msg_buffer_size_bytes` - How big of a message buffer to use for sending/receiving messages. A minimum of 1024 bytes is recommended.
        """
        self.__msg_buffer_size_bytes = config['msg_buffer_size_bytes']

        assert (self.__create_connection(
            ip=config['server_ip'], json_msg_port=config['json_msg_port'], bin_msg_port=config['bin_msg_port']))

        self.__num_channels = self.read_general_info()['TestChannels']

    def get_num_channels(self) -> int:
        '''
        Returns the number of channels associated with the cycler
        '''
        return self.__num_channels

    def read_system_info(self) -> dict:
        """
        Reads the system info from the cycler.  MacNet message (1,1)

        Returns
        -------
        system_info : dict
            The system information for the cycler.
        """
        msg_outgoing_dict = copy.deepcopy(pymacnet.messages.tx_system_info_msg)

        rx_msg = self._send_receive_json_msg(msg_outgoing_dict)
        if rx_msg:
            return rx_msg['result']
        else:
            logger.error("Failed to read system info!")
            return None

    def read_general_info(self) -> dict:
        """
        Reads the general system info from the cycler.  MacNet message (1,2)

        Returns
        -------
        general_info : dict
            The general information for the cycler.
        """
        msg_outgoing_dict = copy.deepcopy(
            pymacnet.messages.tx_general_info_msg)

        rx_msg = self._send_receive_json_msg(msg_outgoing_dict)
        if rx_msg:
            return rx_msg['result']
        else:
            logger.error("Failed to read system info!")
            return None

    def read_channel_status(self, channel: int) -> dict:
        """
        Reads channel status for the specified `channel`.

        Returns
        -------
        status : dict
            A dictionary detailing the status of the channel. Returns None if there is an issue.
        """
        status = {}

        if not (channel > 0) and not (channel < self.__num_channels):
            logger.warning("Invalid channel number!")
            return None

        try:
            msg_outgoing_dict = copy.deepcopy(pymacnet.messages.tx_read_status_msg)
            msg_outgoing_dict['params']['Chan'] = channel
            status = self._send_receive_json_msg(msg_outgoing_dict)
        except Exception as e:
            logger.error(
                f'Error reading channel status for channel {channel}', exc_info=True)
            logger.error(e)

        if status and 'result' in status:
            return status['result']
        else:
            logger.error("Failed to read channel status")
            return None

    def read_all_channel_statuses(self) -> list:
        """
        Reads the channel status for all channels on the cycler.

        Returns
        -------
        statuses : list
            A list of channel statues where the index corresponds to the channel number (zero indexed.)
        """
        msg_outgoing_dict = copy.deepcopy(
            pymacnet.messages.tx_channel_status_multiple_channels)

        msg_outgoing_dict['params']['Len'] = self.__num_channels

        rx_msg = self._send_receive_json_msg(msg_outgoing_dict)
        if rx_msg:
            return rx_msg['result']['Status']
        else:
            logger.error("Failed to read channel status")
            return None

    def _send_receive_json_msg(self, outgoing_msg_dict) -> dict:
        """
        Sends and receives a JSON message to/from the Maccor server.

        Parameters
        ----------
        msg_outgoing_dict : dict
            A dictionary containing the message to be sent.

        Returns
        ----------
        msg_incoming_dict : dict
            A dictionary containing the message response. Returns None if there is an issue.
        """

        if self.__json_msg_socket:
            pass
        else:
            logger.error(
                "__json_msg_socket connection does not exist!", exc_info=True)
            return None

        # Take care of channel zero indexing on outgoing messages
        if 'params' in outgoing_msg_dict and 'Chan' in outgoing_msg_dict['params']:
            outgoing_msg_dict['params']['Chan'] -= 1

        try:
            msg_outgoing_packed = json.dumps(outgoing_msg_dict, indent=4)
            msg_outgoing_packed = msg_outgoing_packed.encode()
        except:
            logger.error("Error packing outgoing message!", exc_info=True)
            return None

        try:
            self.__json_msg_socket.send(msg_outgoing_packed)
        except:
            logger.error("Error sending message!", exc_info=True)
            return None

        try:
            msg_incoming_packed = b''
            while msg_incoming_packed[-2:] != b'\r\n':
                msg_incoming_packed += self.__json_msg_socket.recv(
                    self.__msg_buffer_size_bytes)
        except socket.timeout:
            logger.error(
                "Timeout on receiving message from Maccor server!", exc_info=True)
            return False
        except:
            logger.error("Error receiving message!", exc_info=True)
            return None

        try:
            msg_incoming_dict = json.loads(msg_incoming_packed.decode('utf-8'))
        except:
            logger.error("Error unpacking incoming message!", exc_info=True)
            logger.error("Message: " + str(msg_incoming_packed))
            return None

        # Take care of channel zero indexing on incoming messages
        if 'result' in msg_incoming_dict and 'Chan' in msg_incoming_dict['result']:
            msg_incoming_dict['result']['Chan'] += 1

        return msg_incoming_dict

    def _send_direct_output_rest_bin_msg(self, direct_out_msg_dict: dict) -> bool:
        """
        Sends a direct output binary message to set the channel to rest.
        Note this is necessary because there is a bug where direct output 
        JSON messages cannot set rest.

        Parameters
        ----------
        direct_out_msg_dict : dict
            A direct output message dictionary

        Returns
        -------
        success : bool
            True of False based on whether or not rest was set.
        """

        # Take care of channel zero indexing
        direct_out_msg_dict["params"]['Chan'] += 1

        msg_outgoing_bytes = struct.pack('<HHHHffffBB',
                                         direct_out_msg_dict["params"]['FClass'],
                                         direct_out_msg_dict["params"]['FNum'],
                                         direct_out_msg_dict["params"]['Chan'],
                                         18,
                                         direct_out_msg_dict["params"]['Current'],
                                         direct_out_msg_dict["params"]['Voltage'],
                                         direct_out_msg_dict["params"]['Power'],
                                         direct_out_msg_dict["params"]['Resistance'],
                                         direct_out_msg_dict["params"]['CurrentRange'],
                                         ord('R'))
        try:
            self.__bin_msg_socket.send(msg_outgoing_bytes)
        except:
            logger.error("Error sending binary rest message!", exc_info=True)
            return False
        try:
            response = self.__bin_msg_socket.recv(self.__msg_buffer_size_bytes)
        except socket.timeout:
            logger.error(
                "Timeout on receiving message from Maccor server!", exc_info=True)
            return False
        except:
            logger.error(
                "Error receiving rest message response!", exc_info=True)
            return False
        if response:
            # TODO: Should check something related to the response.
            return True
        else:
            logger.error("No response for setting rest!")
            return False

    def __create_connection(self, ip: str, json_msg_port: int, bin_msg_port: int) -> bool:
        """
        Creates a connection with Maccor server to send/receive JSON and binary messages.

        Parameters
        ----------
        ip : str
            The IP address of the Maccor server.
        json_msg_port : int
            The TCP port use for JSON message communication.
        bin_msg_port : 
            The TCP port used for binary message communication

        Returns
        ----------
        success : bool
            True or False based on whether the connection was created successfully
        """
        try:
            self.__json_msg_socket = socket.socket(
                socket.AF_INET, socket.SOCK_STREAM)
            self.__json_msg_socket.settimeout(2)
            self.__json_msg_socket.connect(
                (ip, json_msg_port)
            )
        except:
            logger.error(
                "Failed to create JSON message socket!", exc_info=True)
            return False
        try:
            self.__bin_msg_socket = socket.socket(
                socket.AF_INET, socket.SOCK_STREAM)
            self.__bin_msg_socket.settimeout(2)
            self.__bin_msg_socket.connect(
                (ip, bin_msg_port)
            )
        except:
            logger.error(
                "Failed to create binary message socket!", exc_info=True)
            return False

        return True



import socket
import json
import threading
import copy
import pymacnet.messages


class ChannelData:

    __chan_status_list = []
    __chan_status_lock = threading.Lock()

    def __init__(self, num_channels):
        """
        Container class that will hold all of the specific channel data for MaccorSpoofer.

        Parameters
        ----------
            num_channels : int
                Number of channels in our hypothetical Maccor cycler.
        """
        self.num_channels = num_channels

        # Create status entries for all of the channels.
        for i in range(0, self.num_channels):
            channel_status = copy.deepcopy(
                pymacnet.messages.rx_read_status_msg
            )
            channel_status['result']['Chan'] = i
            with self.__chan_status_lock:
                self.__chan_status_list.append(copy.deepcopy(channel_status))

    def fetch_channel_status(self, channel):
        """
        Returns the status message for a specified channel.

        Parameters
        ----------
        channel : int
            The channel to return the status for.

        Returns
        -------
        status : dict
            The status message for the requested channel. Empty if channel larger than number of channel
        """
        if channel > self.num_channels:
            return {}
        else:
            with self.__chan_status_lock:
                return copy.deepcopy(self.__chan_status_list[channel])

    def update_channel_status(self, channel, updated_status):
        """
        Updates the stored channel status for the specified channel.

        Parameters
        ----------
        channel : int
            The channel to update the status for.
        updated_status : dict
            Complete or partial dictionary of status values to update.

        Returns
        -------
        success : bool
            Returns True if all values in the updated_status were used to update the channel_status_array.
        """

        if channel > self.num_channels:
            return False
        else:
            for key in updated_status.keys():
                if key not in pymacnet.messages.rx_read_status_msg['result'].keys():
                    return False

            with self.__chan_status_lock:
                for key in updated_status.keys():
                    self.__chan_status_list[channel]['result'][key] = updated_status[key]

        return True


class MaccorSpoofer:

    __client_connect_timeout_s = 1
    __stop_servers_lock = threading.Lock()
    __stop_servers = False

    def __init__(self, config: dict):
        """
        Class to mimic behavior of Maccor cycler MacNet control server. The class is currently dumb 
        and just sends back basic response messages without any notion of channel status or readings. 
        It could be expanded in future. 

        Parameters
        ----------
        config : dict
            A configuration dictionary containing the server ip address and ports to use. The
            fields to include are as follows:

            `server_ip`: The server IP address to host from. Most often 'localhost' for testing.

            `json_port`: The port to use for the JSON server.

            `tcp_port`: the port to use for the TCP server.

            `num_channels`: The number of channel our fictional cycler has.
        """

        self.__channel_data = ChannelData(config['num_channels'])

        # Start a dedicated server thread for processing JSON messages
        json_server_config = {
            'ip': config['server_ip'], 'port': config['json_port']
        }
        self.__json_server_thread = threading.Thread(
            target=self.__server_loop,
            args=(json_server_config, _JsonWorker,),
            daemon=True
        )

        # Start a dedicated server thread for processing binary messages
        tcp_server_config = {
            'ip': config['server_ip'], 'port': config['tcp_port']
        }
        self.__tcp_server_thread = threading.Thread(
            target=self.__server_loop,
            args=(tcp_server_config, _TcpWorker,),
            daemon=True
        )

    def start(self):
        """
        Starts the server loops.
        """
        self.__json_server_thread.start()
        self.__tcp_server_thread.start()

    def update_channel_status(self, channel, updated_status):
        """
        Updates the stored channel status for the specified channel.

        Parameters
        ----------
        channel : int
            The channel to update the status for.
        updated_status : dict
            Complete or partial dictionary of status values to update.

        Returns
        -------
        success : bool
            Returns True if all values in the updated_status were used to update the channel_status_array.
        """
        return self.__channel_data.update_channel_status(channel, updated_status)

    def __server_loop(self, sock_config: dict, Worker):
        """
        Creates a server and forever loop to service client socket requests.

        Parameters
        ----------
        sock_config : dict
            A configuration for the socket containing the IP and port number.
        Worker : _SocketWorker
            A reference to the worker class that will service individual client connections.
        """
        # List that will hold all the workers to service client connections.
        client_workers = []

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((sock_config["ip"], sock_config["port"]))
        sock.settimeout(self.__client_connect_timeout_s)
        sock.listen()

        while True:
            try:
                client_connection = sock.accept()[0]
                client_workers.append(
                    Worker(client_connection, self.__channel_data))
            except socket.timeout:
                with self.__stop_servers_lock:
                    # If stop command is issued then kill all workers.
                    if self.__stop_servers:
                        for worker in client_workers:
                            if worker.is_alive():
                                worker.kill_worker()
                        break
                # Remove any workers that made have died from disconnecting clients.
                client_workers[:] = [
                    worker for worker in client_workers if worker.is_alive()
                ]

    def stop(self):
        """
        Stop the server loops.
        """
        with self.__stop_servers_lock:
            self.__stop_servers = True
        self.__json_server_thread.join()
        self.__tcp_server_thread.join()

    def __del__(self):
        self.stop()


class _SocketWorker:
    """
    Generic worker class that will respond to client socket requests. 
    Default setup as an echo server. Child classes should overwrite the 
    the `_process_client_msg()` method with their own responses.
    """
    __receive_msg_timeout_s = 0.5
    __msg_buffer_size_bytes = 4096
    __stop_lock = threading.Lock()
    __stop = False

    def __init__(self, s: socket.socket):
        """
        Creates the thread to service client requests.

        Parameters
        ----------
        s : socket.socket
            Socket connection to client.
        """
        self.stop = False
        self.__client_thread = threading.Thread(
            target=self.___service_loop,
            args=(s,),
            daemon=True
        )
        self.__client_thread.start()

    def ___service_loop(self, s: socket.socket):
        """
        Forever loop to service client requests. Wait to receive a message. If no messages is 
        received before the timeout then check to see if stop command has been issued. Loop is 
        also broken if client breaks connection by sending b''. 

        Parameters
        ----------
        s : socket.socket
            Socket connection to client.
        """
        s.settimeout(self.__receive_msg_timeout_s)

        while True:
            try:
                rx_msg = s.recv(self.__msg_buffer_size_bytes)
                if not rx_msg:
                    break

                tx_msg = self._process_client_msg(rx_msg)

                s.sendall(tx_msg)
            except socket.timeout:
                with self.__stop_lock:
                    if self.__stop:
                        break
        s.close()

    def _process_client_msg(self, rx_msg):
        """
        Takes the incoming client message and generates a response. 

        Parameters
        ----------
        rx_msg : PyBytesObject
            The client message received.

        Returns
        -------
        tx_msg : PyBytesObject
            The client response.
        """
        return rx_msg

    def is_alive(self):
        """
        Method to call to see if the client service thread is still running.

        Returns
        -------
        running : bool
            True of False based on whether or not the client thread is running.
        """
        return self.__client_thread.is_alive()

    def kill_worker(self):
        """
        Method to stop client service loop.
        """
        if self.__client_thread.is_alive():
            with self.__stop_lock:
                self.__stop = True
            self.__client_thread.join()


class _JsonWorker(_SocketWorker):

    def __init__(self, s: socket.socket, channel_data: ChannelData):
        """
        Class to handle requests from MacNet JSON socket clients.

        Parameters
        ----------
        s : socket.socket
            Socket to communicate with.
        channel_data : ChannelData
            Container class of channel data
        """

        # These must be created before calling ().__init__(s). Otherwise access can be attempted before they exist.
        self.__channel_data = channel_data
        super().__init__(s)

    def _process_client_msg(self, rx_msg):
        """
        Takes the incoming JSON client message and generates a response.

        Parameters
        ----------
        rx_msg : PyBytesObject
            The client message received.

        Returns
        -------
        tx_msg : PyBytesObject
            The client response.
        """

        rx_msg = json.loads(rx_msg)

        if (pymacnet.messages.tx_read_status_msg['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_read_status_msg['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = copy.deepcopy(
                self.__channel_data.fetch_channel_status(
                    rx_msg['params']['Chan'])
            )
        elif (pymacnet.messages.tx_read_aux_msg['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_read_aux_msg['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = pymacnet.messages.rx_read_aux_msg
            tx_msg['result']['Chan'] = rx_msg['params']['Chan']
        elif (pymacnet.messages.tx_start_test_with_procedure_msg['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_start_test_with_procedure_msg['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = pymacnet.messages.rx_start_test_with_procedure_msg
            tx_msg['result']['Chan'] = rx_msg['params']['Chan']
        elif (pymacnet.messages.tx_set_variable_msg['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_set_variable_msg['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = pymacnet.messages.rx_set_variable_msg
            tx_msg['result']['Chan'] = rx_msg['params']['Chan']
        elif (pymacnet.messages.tx_start_test_with_direct_control_msg['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_start_test_with_direct_control_msg['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = pymacnet.messages.rx_start_test_with_direct_control_msg
            tx_msg['result']['Chan'] = rx_msg['params']['Chan']
        elif (pymacnet.messages.tx_set_direct_output_msg['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_set_direct_output_msg['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = pymacnet.messages.rx_set_direct_output_msg
            tx_msg['result']['Chan'] = rx_msg['params']['Chan']
        elif (pymacnet.messages.tx_reset_channel_msg['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_reset_channel_msg['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = pymacnet.messages.rx_reset_channel_msg
            tx_msg['result']['Chan'] = rx_msg['params']['Chan']
        elif (pymacnet.messages.tx_set_safety_limits_msg['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_set_safety_limits_msg['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = pymacnet.messages.rx_set_safety_limits_msg
            tx_msg['result']['Chan'] = rx_msg['params']['Chan']
            tx_msg['result']['VSafeMax'] = rx_msg['params']['VSafeMax']
            tx_msg['result']['VSafeMin'] = rx_msg['params']['VSafeMin']
            tx_msg['result']['ISafeChg'] = rx_msg['params']['ISafeChg']
            tx_msg['result']['ISafeDis'] = rx_msg['params']['ISafeDis']
            tx_msg['result']['PBatSafeChg'] = rx_msg['params']['PBatSafeChg']
            tx_msg['result']['PBatSafeDis'] = rx_msg['params']['PBatSafeDis']
        elif (pymacnet.messages.tx_system_info_msg['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_system_info_msg['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = pymacnet.messages.rx_system_info_msg
        elif (pymacnet.messages.tx_general_info_msg['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_general_info_msg['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = pymacnet.messages.rx_general_info_msg
            tx_msg['result']['TestChannels'] = self.__channel_data.num_channels
        elif (pymacnet.messages.tx_channel_status_multiple_channels['params']['FClass'] == rx_msg['params']['FClass'] and
                pymacnet.messages.tx_channel_status_multiple_channels['params']['FNum'] == rx_msg['params']['FNum']):
            tx_msg = pymacnet.messages.rx_channel_status_multiple_channels
        else:
            tx_msg = {'err': 1}

        tx_msg = json.dumps(tx_msg, indent=4)
        # Needs to be included as it's included in messages from Maccor server as indicator of message termination
        tx_msg += '\r\n'
        tx_msg = tx_msg.encode('utf-8')

        return tx_msg


class _TcpWorker(_SocketWorker):

    def __init__(self, s: socket.socket, channel_data: ChannelData):
        """
        Class to handle requests from MacNet TCP socket clients. Currently just echos back client
        message. Should be expanded in future.

        Parameters
        ----------
        s : socket.socket
            Socket to communicate with.
        channel_data : ChannelData
            Container class of channel data
        """

        self.__channel_data = channel_data
        super().__init__(s)



import setuptools

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pymacnet",
    version="1.1.1",
    author="Zander Nevitt",
    author_email="zandern@battgenie.life",
    description="A class based python interface for communication and control of Maccor cyclers over Macnet.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BattGenie/pymacnet.git",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    license='MIT',
)


status_dictionary = {
    0: 'Available',
    1: 'Selected',
    2: 'Active',
    3: 'Suspended',
    4: 'Completed',
    5: 'Problem',
    6: 'Not/Avl',
    7: 'Reset',
    8: 'Start',
    9: 'PWR Fail',
    10: 'No Cntllr',
    11: 'ShutDown',
    12: 'Starting',
    13: 'Blocked',
    14: 'Waiting',
    15: 'FRA',
    16: 'Rem Maint'
}
"""
Dictionary to decode the "result" field returned by the channel status messages.
"""


tx_read_status_msg = {
    'jsonrpc': '2.0',
    'method': 'MacNet',
    'params':
    {
        'FClass': 4,
        'FNum': 7,
        'Chan': -1
    },
    'id': 1987
}
"""
Gets the status and readings of one channel. The status and readings of `Chan` channel will be returned.
"""

rx_read_status_msg = {
    'jsonrpc': '2.0',
    'result':
    {
        'FClass': 4,
        'FNum': 7,
        'Chan': -1,
        'RF1': 0,
        'RF2': 192,
        'Stat': 0,
        'LastRecNum': 4225,
        'Cycle': 0,
        'Step': 5,
        'TestTime': 2.00,
        'StepTime': 1.00,
        'Capacity': 0,
        'Energy': 0,
        'Current': 0,
        'Voltage': 3.85,
        'TesterTime': '2022-10-13T12:32:56'
    },
    'id': 1987
}
"""
All status and readings from channel `Chan`.
"""

tx_start_test_with_procedure_msg = {
    "jsonrpc": "2.0",
    "method": "MacNet",
    "params":
    {
        "FClass": 6,
        "FNum": 2,
        "Chan": -1,
        "TestName": "Random",
        "ProcName": "Procedure Name",
        "Comment": "Test comment",
        "Crate": 1,
        "ChamberNum": 0,
        "StartCycle": 0,
        "TotCycles": 0,
        "Mass": 1,
        "VGain": 0,
        "AbsTRepAlign": 0,
        "ParallelR": 0,
        "VDivHiR": 0,
        "VDivLoR": 0,
        "CANpos": -1,
        "CANprof": "",
        "RegimeName": ""
    },
    "id": 1987
}
"""
Start test with direct control

Selects the test channel to be started. The test name and procedure name can be up to 250 characters long.
Also, do not select the channel in advance of starting the test. The JSON Start test command will both select and 
start the channel. If, however, multiple channels should be started with the same test, do first select the channels,
then start the test using 65535, meaning "all selected".

Parameters as follows:

`TestName` Up to 25 ascii characters for the data file name. “Random” will generate a pseudo-random name. 
The field is fixed in length, so remaining characters should be space characters.

`ProcName` Up to 25 ascii characters for the test procedure name. The test procedure must exist in the 
C:\\Maccor\\Procedure folder and the “.000” should not be part of the name. The field is fixed in length, 
so remaining characters should be space characters.

`Comment` Up to 80 characters for the test comment. The field is fixed in length, 
so remaining characters should be space characters.

`Crate` C-rate. If C-rate is not used, enter 1.

`ChamberNum` Environmental chamber number. If the environmental chamber is not used, enter 0.

`StartCycle` Default: 0; See Start Test Setup for further details

`TotCycles` Default: 0; See Start Test Setup for further details

`Mass` Default: 1; See Start Test Setup for further details

`VGain` Default: 0; Used to change the gain of the constant voltage feedback loop. Only use it after consulting Customer Service.
0: Gain x1 1: Gain x2 2: Gain x4 3: Gain x8 4: Gain /1 5: Gain /2 6: Gain /4 7: Gain /8

`AbsTRepAlign` Default: 0; 1: Use absolute time report alignment. See Start Test Setup for further details

`ParallelR` See Start Test Setup for further details

`VDivHiR` See Start Test Setup for further details

`VDivLoR` See Start Test Setup for further details
"""

rx_start_test_with_procedure_msg = {
    'jsonrpc': '2.0',
    'result':
    {
        'FClass': 6,
        'FNum': 2,
        'Chan': -1,
        'Result': 'OK'
    },
    'id': 1987
}
"""
Response for starting test with procedure.
"""

tx_start_test_with_direct_control_msg = {
    "jsonrpc": "2.0",
    "method": "MacNet",
    "params":
    {
        "FClass": 6,
        "FNum": 7,
        "Chan": -1,
        "TestName": "Random",
        "Current": 0.0,
        "Voltage": 5.0,
        "Power": 50,
        "Resistance": 0,
        "CurrentRange": 1,
        "ChMode": 'D',
        "DataTime": 1.0,
        "DataV": 0,
        "DataI": 0
    },
    "id": 1987
}
"""
Start direct mode test.

Start direct mode. Initializes direct remote control of the test channel with the 
following data in the “Data” area. The channel will output this current, voltage 
and power - mutually limiting as standard - until the values are changed with the 
“(6, 8) Set direct output” command described below or stopped.

A watchdog timer can be activated in the Misc_Options section of the System.ini file. Setting the 
MacNetDirectModeWD to a value higher than 0 will cause a channel under direct mode to suspend if it 
has not been updated within the specified number of seconds.

Parameters are as follows:

`testName` Up to 25 ascii characters for the data file name. “Random” (Case sensitive) will 
generate a pseudo-random name. The field is fixed in length, so remaining characters should be space characters.

`Current` Amps. Mutually limiting. Set outside range to ignore. 0 is within range. 
Must be active to function in (6, 8) Set direct mode output.

`Voltage` Voltage. Mutually limiting. Set outside range to ignore. 0 is within range. 
Must be active to function in (6, 8) Set direct mode output.

`Power` Watts. Mutually limiting. Set outside range to ignore. 0 is within range. 
Must be active to function in (6, 8) Set direct mode output.

`Resistance`  Ohms. Mutually limiting. Set to 0 to ignore. 
Must be active to function in (6, 8) Set direct mode output.

`CurrentRange` The desired current range: 1, 2, 3, 4

`ChMode` Charge mode. 67 i.e. 'C' for charge, 68 i.e. 'D' for discharge and 82 i.e. 'R' for rest

`DataTime` Data record time. Time increment between data records in the data file. Enter 0 for no data.
Data record voltage

`DataV` `Voltage increment to generate a data record in the data file. Enter 0 to disable.
Data record current

`DataI` Current increment to generate a data record in the data file. Enter 0 to disable.
"""

rx_start_test_with_direct_control_msg = {
    'jsonrpc': '2.0',
    'result':
    {
        'FClass': 6,
        'FNum': 7,
        'Chan': -1,
        'Result': 'OK'
    },
    'id': 1987
}
"""
Response message for starting test with direct control.
"""

tx_set_direct_output_msg = {
    "jsonrpc": "2.0",
    "method": "MacNet",
    "params":
    {
        "FClass": 6,
        "FNum": 8,
        "Chan": -1,
        "Current": 0.0,
        "Voltage": 65536,
        "Power": 65536,
        "Resistance": 0,
        "CurrentRange": 4,
        "ChMode": "C"
    },
    "id": 1987
}
"""
Set direct mode output

Set direct output. When a channel has been started in direct mode, the output can be set with this 
command and these arguments:

Parameters as follows:

`Current` Amps. Mutually limiting. Set outside range to ignore. 0 is within range. 
        Must be active to function in (6, 8) Set direct mode output.

`Voltage` Voltage. Mutually limiting. Set outside range to ignore. 0 is within range. 
Must be active to function in (6, 8) Set direct mode output.

`Power` Watts. Mutually limiting. Set outside range to ignore. 0 is within range. 
Must be active to function in (6, 8) Set direct mode output.

`Resistance` Ohms. Mutually limiting. Set to 0 to ignore. 
Must be active to function in (6, 8) Set direct mode output.

`Current range` The desired current range: 1, 2, 3, 4

`Charge mode` 67 i.e. 'C' for charge, 68 i.e. 'D' for discharge and 82 i.e. 'R' for rest
Note: As of 12/14/22 setting 'R' for rest does not work and rest must be set via other means.

Response back results decode as follows:
    0: OK
    1: Illegal system type. This feature only works for type 0
    2: The channel is not active
    3: Command sent too fast. There must be at least 100 ms (10 ticks) between commands.
    4: Direct mode is not active.
    5: Direct mode is not ready yet.
"""

rx_set_direct_output_msg = {
    'jsonrpc': '2.0',
    'result':
    {
        'FClass': 6,
        'FNum': 8,
        'Chan': -1,
        'Result': 'OK'
    },
    'id': 1987
}
"""
Response for setting direct mode output.
"""

tx_reset_channel_msg = {
    "jsonrpc": "2.0",
    "method": "MacNet",
    "params":
    {
        "FClass": 6,
        "FNum": 5,
        "Chan": -1
    },
    "id": 1987
}
"""
Reset “Chan” test channel.
"""

rx_reset_channel_msg = {
    "jsonrpc": "2.0",
    "result":
    {
        "FClass": 6,
        "FNum": 5,
        "Chan": -1,
        "Result": "OK"
    },
    "id": 1987
}
"""
Reset channel response.
"""

tx_read_aux_msg = {
    "jsonrpc": "2.0",
    "method": "MacNet",
    "params":
    {
        "FClass": 4,
        "FNum": 4,
        "Chan": -1
    },
    "id": 1987
}
"""
Get the auxiliary values for the specified channel `Chan`.
"""

rx_read_aux_msg = {
    'jsonrpc': '2.0',
    'result':
    {
        'FClass': 4,
        'FNum': 4,
        'Chan': -1,
        'Len': 1,
        'AuxValues': [24.75]
    },
    'id': 1987
}
"""
The auxiliary response message.
"""

tx_set_safety_limits_msg = {
    "jsonrpc": "2.0",
    "method": "MacNet",
    "params":
    {
        "FClass": 6,
        "FNum": 10,
        "Chan": 3,
        "VSafeMax": 4.5,
        "VSafeMin": 2.5,
        "ISafeChg": 4.3,
        "ISafeDis": 2.5,
        "PBatSafeChg": 25,
        "PBatSafeDis": 25
    },
    "id": 1987
}
"""
Sets safety limits on a channel. (6, 10)
"""

rx_set_safety_limits_msg = {
    "jsonrpc": "2.0",
    "result":
    {
        "FClass": 6,
        "FNum": 10,
        "Chan": -1,
        "VSafeMax": 0,
        "VSafeMin": 0,
        "ISafeChg": 0,
        "ISafeDis": 0,
        "PBatSafeChg": 0,
        "PBatSafeDis": 0
    },
    "id": 1987
}
"""
Response message for setting safety limits on channel. (6, 10)
"""

tx_set_variable_msg = {
    "jsonrpc": "2.0",
    "method": "MacNet",
    "params":
    {
        "FClass": 6,
        "FNum": 9,
        "Chan": -1,
        "VarNum": 1,
        "Value": 0
    },
    "id": 1987
}
"""
Sets channel variable for the specified variable 'VarNum'.
"""

rx_set_variable_msg = {
    'jsonrpc': '2.0',
    'result':
    {
        'FClass': 6,
        'FNum': 9,
        'Chan': -1,
        'Result': 'OK'
    },
    'id': 1987
}
"""
Response message for setting channel variables.
"""

tx_system_info_msg = {
    "jsonrpc": "2.0", 
    "method": "MacNet", 
    "params":
    {
    "FClass": 1,
    "FNum": 1 
    },
    "id": 1987 
}
"""
Requests the software system information from the Maccor cycler, message (1,1)
"""

rx_system_info_msg = {
    "jsonrpc":"2.0", 
    "result":
    {
        "FClass":1,
        "FNum":1,
        "APIVersion":1, 
        "MacTest32EXEversionMajor":3, 
        "MacTest32EXEversionMinor":2, 
        "MacTest32EXEversionBuild":18, 
        "MacTest32DLLversionMajor":3, 
        "MacTest32DLLversionMsinor":2, 
        "MacTest32DLLversionBuild":18, 
        "MacTest32ExeDT":"2016-11-08T15:02:58", 
        "MacTest32DLLDT":"2016-11-13T11:29:48"
    },
    "id":1987
}
"""
Example response for system software version, message (1,1)
"""

tx_general_info_msg = {
    "jsonrpc": "2.0", 
    "method": "MacNet", 
    "params":
    {
        "FClass": 1,
        "FNum": 2 
    },
    "id": 1987 
}
"""
Requests the general system information from the Maccor cycler, message (1,2)
"""

rx_general_info_msg = {
    "jsonrpc":"2.0", 
    "result":
    {
        "FClass":1, 
        "FNum":2, 
        "SystemID":"Win10", 
        "SystemType":0, 
        "ControllerBoards":3, 
        "TestChannels":12, 
        "AuxBoards":1, 
        "AuxChannels":128, 
        "SMB1Pos":0,
        "SMB3Pos":1 
    },
    "id":1987
}
"""
Example response for general system information from the Maccor cycler, message (1,2)
"""

tx_channel_status_multiple_channels = {
    "jsonrpc": "2.0", 
    "method": "MacNet", 
    "params":
    {
        "FClass": 4, 
        "FNum": 1, 
        "Chan": 0, 
        "Len": 2
    },
    "id": 1987
}
"""
Requests the status of “Len” channels starting from “Chan” will be returned. 
Four bytes per requested channel. Up to 128 channels can be requested in one message.

MacNet message: (4, 1) Channel status of multiple channels
"""

rx_channel_status_multiple_channels = {
    "jsonrpc":"2.0",
    "result":
    {
        "FClass":4, 
        "FNum":1, 
        "Chan":0, 
        "Len":2, 
        "Status":
        [
            {
                "RF1":0, 
                "RF2":0, 
                "Stat":0
            }, 
            {
                "RF1":0,
                "RF2":0, 
                "Stat":0
            } 
        ]
    }, 
    "id":1987
}
"""
Example response for MacNet message: (4, 1) Channel status of multiple channels)
"""